#ifndef _L4_VREGS_H
# error "Never use <l4/bits/vregs.h> directly; include <l4/vregs.h> instead."
#endif

#ifndef _L4_EXTERN_INLINE
#define _L4_EXTERN_INLINE extern __inline
#endif

/* Retrieve the UTCB address.  */
static inline l4_word_t *
__l4_utcb (void)
{
  l4_word_t *utcb;

  __asm__ ("movl %%gs:0, %[utcb]"
	   : [utcb] "=r" (utcb));

  return utcb;
}

/* Offsets of various elements in the UTCB, relativ to the UTCB
   address.  */
#define __L4_UTCB_MR0		0
#define __L4_UTCB_THREAD_WORD0	-4
#define __L4_UTCB_THREAD_WORD1	-5
#define __L4_UTCB_SENDER	-6
#define __L4_UTCB_RECEIVER	-7
#define __L4_UTCB_TIMEOUT	-8
#define __L4_UTCB_ERROR_CODE	-9
#define __L4_UTCB_FLAGS		-10
#define __L4_UTCB_EXC_HANDLER	-11
#define __L4_UTCB_PAGER		-12
#define __L4_UTCB_USER_HANDLE	-13
#define __L4_UTCB_PROCESSOR_NO	-14
#define __L4_UTCB_MY_GLOBAL_ID	-15
#define __L4_UTCB_BR0		-16


/* Get the local thread ID.  */
_L4_EXTERN_INLINE l4_thread_id_t
l4_my_local_id (void)
{
  l4_thread_id_t id;

  /* Local thread ID is equal to the UTCB address.  */
  id.raw = (l4_word_t) __l4_utcb ();

  return id;
}


/* Get the global thread ID.  */
_L4_EXTERN_INLINE l4_thread_id_t
l4_my_global_id (void)
{
  l4_thread_id_t id;
  l4_word_t *utcb = __l4_utcb ();

  id.raw = utcb[__L4_UTCB_MY_GLOBAL_ID];
  return id;
}


_L4_EXTERN_INLINE int
l4_processor_no (void)
{
  l4_word_t *utcb = __l4_utcb ();

  return utcb[__L4_UTCB_PROCESSOR_NO];
}


_L4_EXTERN_INLINE l4_word_t
l4_user_defined_handle (void)
{
  l4_word_t *utcb = __l4_utcb ();

  return utcb[__L4_UTCB_USER_HANDLE];
}


_L4_EXTERN_INLINE void
l4_set_user_defined_handle (l4_word_t data)
{
  l4_word_t *utcb = __l4_utcb ();

  utcb[__L4_UTCB_USER_HANDLE] = data;
}


_L4_EXTERN_INLINE l4_thread_id_t
l4_pager (void)
{
  l4_thread_id_t thread;
  l4_word_t *utcb = __l4_utcb ();

  thread.raw = utcb[__L4_UTCB_PAGER];
  return thread;
}


_L4_EXTERN_INLINE void
l4_set_pager (l4_thread_id_t thread)
{
  l4_word_t *utcb = __l4_utcb ();

  utcb[__L4_UTCB_PAGER] = thread.raw;
}


_L4_EXTERN_INLINE l4_thread_id_t
l4_exception_handler (void)
{
  l4_thread_id_t thread;
  l4_word_t *utcb = __l4_utcb ();

  thread.raw = utcb[__L4_UTCB_EXC_HANDLER];
  return thread;
}


_L4_EXTERN_INLINE void
l4_set_exception_handler (l4_thread_id_t thread)
{
  l4_word_t *utcb = __l4_utcb ();

  utcb[__L4_UTCB_EXC_HANDLER] = thread.raw;
}


_L4_EXTERN_INLINE void
l4_clear_cop_flag (l4_word_t n)
{
  l4_word_t *utcb = __l4_utcb ();

  /* The second byte in the flags field contains the coprocessor
     flags.  */
  __asm__ ("andl %[mask], %[where]"
	   :
	   : [mask] "ir" (~(1 << (8 + n))),
	   [where] "m" (utcb[__L4_UTCB_FLAGS]));
}


_L4_EXTERN_INLINE void
l4_set_cop_flag (l4_word_t n)
{
  l4_word_t *utcb = __l4_utcb ();

  /* The second byte in the flags field contains the coprocessor
     flags.  */
  __asm__ ("orl %[mask], %[where]"
	   :
	   : [mask] "ir" (1 << (8 + n)),
	   [where] "m" (utcb[__L4_UTCB_FLAGS]));
}


_L4_EXTERN_INLINE l4_word_t
l4_error_code (void)
{
  l4_word_t *utcb = __l4_utcb ();

  return utcb[__L4_UTCB_ERROR_CODE];
}


_L4_EXTERN_INLINE l4_word_t
l4_xfer_timeout (void)
{
  l4_word_t *utcb = __l4_utcb ();

  return utcb[__L4_UTCB_TIMEOUT];
}


_L4_EXTERN_INLINE void
l4_set_xfer_timeout (l4_word_t time)
{
  l4_word_t *utcb = __l4_utcb ();

  utcb[__L4_UTCB_TIMEOUT] = time;
}


_L4_EXTERN_INLINE l4_thread_id_t
l4_intended_receiver (void)
{
  l4_thread_id_t thread;
  l4_word_t *utcb = __l4_utcb ();

  thread.raw = utcb[__L4_UTCB_RECEIVER];
  return thread;
}


_L4_EXTERN_INLINE l4_thread_id_t
l4_actual_sender (void)
{
  l4_thread_id_t thread;
  l4_word_t *utcb = __l4_utcb ();

  thread.raw = utcb[__L4_UTCB_SENDER];
  return thread;
}


_L4_EXTERN_INLINE void
l4_set_virtual_sender (l4_thread_id_t thread)
{
  l4_word_t *utcb = __l4_utcb ();

  utcb[__L4_UTCB_SENDER] = thread.raw;
}


/* Message registers (MR0 to MR63) start at offset __L4_UTCB_MR0 and
   go upward.  The spec doesn't say this, but we can use __L4_UTCB_MR0
   to store the first message register MR0 until we do an actual
   system call, although only MR1 to MR63 are actually memory
   mapped.  */

/* Set message register NR to DATA.  */
_L4_EXTERN_INLINE void
l4_load_mr (int nr, l4_word_t data)
{
  l4_word_t *mr = __l4_utcb () + __L4_UTCB_MR0;

  mr[nr] = data;
}


/* Set COUNT message registers beginning from START to the values in
   DATA.  */
_L4_EXTERN_INLINE void
l4_load_mrs (int start, int count, l4_word_t *data)
{
  l4_word_t *mr = __l4_utcb () + __L4_UTCB_MR0 + start;

  while (count--)
    *(mr++) = *(data++);
}


/* Store message register NR in DATA.  */
_L4_EXTERN_INLINE void
l4_store_mr (int nr, l4_word_t *data)
{
  l4_word_t *mr = __l4_utcb () + __L4_UTCB_MR0;

  *data = mr[nr];
}


/* Store COUNT message registers beginning from START in DATA.  */
_L4_EXTERN_INLINE void
l4_store_mrs (int start, int count, l4_word_t *data)
{
  l4_word_t *mr = __l4_utcb () + __L4_UTCB_MR0 + start;

  while (count--)
    *(data++) = *(mr++);
}


/* Buffer registers (BR0 to BR31) start at offset __L4_UTCB_BR0 and go
   downward.  */

/* Set message register NR to DATA.  */
_L4_EXTERN_INLINE void
l4_load_br (int nr, l4_word_t data)
{
  l4_word_t *br = __l4_utcb () + __L4_UTCB_BR0;

  br[nr] = data;
}


/* Set COUNT message registers beginning from START to the values in
   DATA.  */
_L4_EXTERN_INLINE void
l4_load_brs (int start, int count, l4_word_t *data)
{
  l4_word_t *br = __l4_utcb () + __L4_UTCB_BR0 - start;

  while (count--)
    *(br--) = *(data--);
}


/* Store message register NR in DATA.  */
_L4_EXTERN_INLINE void
l4_store_br (int nr, l4_word_t *data)
{
  l4_word_t *br = __l4_utcb () + __L4_UTCB_BR0;

  *data = br[nr];
}


/* Store COUNT message registers beginning from START in DATA.  */
_L4_EXTERN_INLINE void
l4_store_brs (int start, int count, l4_word_t *data)
{
  l4_word_t *br = __l4_utcb () + __L4_UTCB_BR0 - start;

  while (count--)
    *(data--) = *(br--);
}

