#ifndef _L4_TYPES_H
#define _L4_TYPES_H	1

#define L4_LITTLE_ENDIAN	0
#define L4_BIG_ENDIAN		1

#define L4_WORDSIZE_32		0
#define L4_WORDSIZE_64		1

/* <l4/bits/types.h> defines L4_BYTE_ORDER and L4_WORDSIZE.  It also
   defines the basic type l4_word_t.  */
#include <l4/bits/types.h>

# define __l4_intN_t(N, MODE) \
  typedef int l4_int##N##_t __attribute__ ((__mode__ (MODE)))
# define __l4_uintN_t(N, MODE) \
  typedef unsigned int l4_uint##N##_t __attribute__ ((__mode__ (MODE)))

__l4_intN_t (8, __QI__);
__l4_intN_t (16, __HI__);
__l4_intN_t (32, __SI__);
__l4_intN_t (64, __DI__);

__l4_uintN_t (8, __QI__);
__l4_uintN_t (16, __HI__);
__l4_uintN_t (32, __SI__);
__l4_uintN_t (64, __DI__);

#if L4_WORDSIZE == L4_WORDSIZE_32
typedef l4_uint32_t l4_word_t;
#else
#if L4_WORDSIZE == L4_WORDSIZE_64
typedef l4_uint64_t l4_word_t;
#else
#error "Unsupported word size."
#endif
#endif


/* Utility macros for structure definitions.  */
#if L4_WORDSIZE == L4_WORDSIZE_32
#define _L4_BITFIELD(name,nr32,nr64) l4_word_t name : nr32
#define _L4_BITFIELD_32(name,nr) l4_word_t name : nr
#define _L4_BITFIELD_64(name,nr)
#else
#if L4_WORDSIZE == L4_WORDSIZE_64
#define _L4_BITFIELD(name,nr32,nr64) l4_word_t name : nr64
#define _L4_BITFIELD_32(name,nr)
#define _L4_BITFIELD_64(name,nr) l4_word_t name : nr
#endif
#endif

#define _L4_WORD_STRUCT(x)	\
  union				\
  {				\
    l4_word_t raw;		\
    struct			\
      x;			\
  }


/* The Thread ID type is here to avoid vregs.h requiring thread.h.  */
typedef union
{
  l4_word_t raw;

  struct
  {
    _L4_BITFIELD (version, 14, 32);
    _L4_BITFIELD (thread_no, 18, 32);
  };
  struct
  {
    l4_word_t _all_zero : 6;
    _L4_BITFIELD (local, 26, 58);
  };

  /* FIXME: Remove this when gcc supports initializers for unnamed
     fields, and change the initializers in <l4/thread.h>.  */
  struct
  {
    _L4_BITFIELD (version, 14, 32);
    _L4_BITFIELD (thread_no, 18, 32);
  } global;
  struct
  {
    l4_word_t _all_zero : 6;
    _L4_BITFIELD (local, 26, 58);
  } local;
} l4_thread_id_t;


typedef union
{
  l4_uint64_t raw;
  struct
  {
    l4_uint32_t low;
    l4_uint32_t high;
  };
} l4_clock_t;


typedef union
{
  l4_uint16_t raw;

  /* This is a time period.  It is 2^e * m usec long.  */
  struct
  {
    l4_uint16_t m : 10;
    l4_uint16_t e : 5;
    l4_uint16_t _zero : 1;
  } period;

  /* This is a time point with validity (2^10 - 1) * 2^e.  */
  struct
  {
    l4_uint16_t m : 10;
    l4_uint16_t c : 1;
    l4_uint16_t e : 4;
    l4_uint16_t _one : 1;
  } point;
} l4_time_t;

#endif	/* l4/types.h */
